#open"syntaxe_commune";;

type debruijn_expr =
   DApp of debruijn_expr * debruijn_expr
 | DAbs of debruijn_expr
 | DVar of int
 | DCst of constante
 | DIf of debruijn_expr * debruijn_expr * debruijn_expr
 | DSet of int * debruijn_expr
 | DLetIn of debruijn_expr list * debruijn_expr
 | DLetRecIn of debruijn_expr list * debruijn_expr
 | DPrim of primitive
;;

exception Pas_trouve of string
;;

let position id env =
  let rec pos_rec n = function
      []       -> raise (Pas_trouve id)
    | s::reste -> if id = s then n else pos_rec (n+1) reste
  in pos_rec 1 env
;;

let traduit_en_debruijn e =
  let rec trad_rec env = function
      App (e,e1)     -> DApp (trad_rec env e, trad_rec env e1)
    | Abs (id,e)     -> DAbs (trad_rec (id::env) e)
    | Var id         -> DVar (position id env)
    | Cst cst        -> DCst cst
    | If (e0,e1,e2)  -> DIf (trad_rec env e0, trad_rec env e1,
                                              trad_rec env e2)
    | Set (id,e)     -> DSet (position id env,trad_rec env e)
    | LetIn (idl,el,e)
                     -> DLetIn (map (trad_rec env) el,
                                trad_rec (idl@env) e)
    | LetRecIn (idl,el,e)
                     -> let env' = idl@env in
                        DLetRecIn (map (trad_rec env') el,
                                   trad_rec env' e)
    | Prim prim      -> DPrim prim
  in
    try trad_rec [] e
    with Pas_trouve id -> failwith ("variable "^id^" non definie")
;;

type asm =
   Applique
 | Lambda
 | Fonction of asm list
 | Retour
 | Accede of int
 | Constante of constante
 | Depile of int
 | Reserve of int
 | Let of int
 | Letrec of int
 | Select of asm list * asm list
 | Modifie of int
 | Primitive of primitive
;;

let traduit_en_asm e =
  let rec trad_rec pile = function
     DApp (e,e1)      -> Applique :: (trad_rec (trad_rec pile e1) e)
   | DAbs e           ->
         (Fonction (rev (Retour :: (trad_rec [Lambda] e)))) :: pile
   | DVar pos         -> (Accede pos) :: pile
   | DCst cst         -> (Constante cst) :: pile
   | DIf (e0,e1,e2)   ->
         (Select (rev (Retour :: (trad_rec [] e1)),
                 (rev (Retour :: (trad_rec [] e2)))))
      :: (trad_rec pile e0)
   | DSet (pos,e)     -> (Modifie pos) :: (trad_rec pile e)
   | DLetIn (el,e)    -> 
       let n = list_length el in
       let pile' = it_list (fun pile e -> (trad_rec pile e)) pile el in
       (Depile n) :: (trad_rec ((Let n) :: pile') e)
   | DLetRecIn (el,e) -> 
          let n = list_length el in
       let pile' = (Reserve n) :: pile in
       let pile'' = it_list (fun pile e -> (trad_rec pile e)) pile' el in
       (Depile n) :: (trad_rec ((Letrec n) :: pile'') e)
   | DPrim prim       -> (Primitive prim) :: pile
  in
    rev (Retour :: (trad_rec [] e))
;;

type seqasm =
   SApplique
 | SLambda
 | SFonction of int
 | SRetour
 | SAccede of int
 | SConstante of constante
 | SDepile of int
 | SReserve of int
 | SLet of int
 | SLetrec of int
 | SSelect of int * int
 | SModifie of int
 | SPrimitive of primitive
;;

let rec taille = function
    (Fonction al)::reste      -> (taille al) + (taille reste) + 1
  | (Select (al1,al2))::reste -> (taille al1) + (taille al2) + (taille reste)+1
  |   a::reste                -> (taille reste) + 1
  |   []                      -> 0
;;

let traduit_en_seqasm al =
  let v = make_vect (taille al) SApplique in
  let rec trad_rec p = function
     []                        -> p
   | Applique::reste           ->
         v.(p) <- SApplique;
         trad_rec (p+1) reste
   | Lambda::reste             ->
         v.(p) <- SLambda;
         trad_rec (p+1) reste
   | (Fonction al)::reste      ->
       	 let p' = trad_rec (p+1) reste in
	 v.(p) <- SFonction p';
         trad_rec p' al
   | Retour::reste             ->
         v.(p) <- SRetour;
         trad_rec (p+1) reste
   | (Accede pos)::reste       ->
         v.(p) <- SAccede pos;
         trad_rec (p+1) reste
   | (Constante cst)::reste    ->
         v.(p) <- SConstante cst;
         trad_rec (p+1) reste
   | (Depile n)::reste         ->
         v.(p) <- SDepile n;
         trad_rec (p+1) reste
   | (Reserve n)::reste        ->
         v.(p) <- SReserve n;
         trad_rec (p+1) reste
   | (Let n)::reste            ->
         v.(p) <- SLet n;
         trad_rec (p+1) reste
   | (Letrec n)::reste         ->
         v.(p) <- SLetrec n;
         trad_rec (p+1) reste
   | (Select (al1,al2))::reste ->
       	 let p' = trad_rec (p+1) reste in
         let p'' = trad_rec p' al1 in
	 v.(p) <- SSelect (p',p'');
         trad_rec p'' al2
   | (Modifie pos)::reste      ->
         v.(p) <- SModifie pos;
         trad_rec (p+1) reste
   | (Primitive prim)::reste   ->
         v.(p) <- SPrimitive prim;
         trad_rec (p+1) reste
  in
    trad_rec 0 al;
    v
;;

let compile_commun e =
   traduit_en_seqasm (traduit_en_asm (traduit_en_debruijn e))
;;


